import io
import os
import pathlib
import unittest
from xml.etree.ElementTree import ParseError

from svgelements import *


class TestElementWrite(unittest.TestCase):

    def test_write(self):
        q = io.StringIO(
            u'''
            <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="360pt" height="360pt" viewBox="0 0 360 360" version="1.1">'
            <g id="mygroup">
            <path style="fill:none;stroke-width:0.004;stroke-linecap:butt;stroke-linejoin:miter;stroke:rgb(0%,0%,0%);stroke-opacity:1;stroke-miterlimit:10;" d="M 0.00199653 0 L 0.00199653 1 M 0.0220052 0 L 0.0220052 1 M 0.042003 0 L 0.042003 1 M 0.0620009 0 L 0.0620009 1 M 0.0819987 0 L 0.0819987 1 M 0.101997 0 L 0.101997 1 M 0.122005 0 L 0.122005 1 M 0.142003 0 L 0.142003 1 M 0.162001 0 L 0.162001 1 M 0.181999 0 L 0.181999 1 M 0.201997 0 L 0.201997 1 M 0.222005 0 L 0.222005 1 M 0.242003 0 L 0.242003 1 M 0.262001 0 L 0.262001 1 M 0.281999 0 L 0.281999 1 M 0.301997 0 L 0.301997 1 M 0.322005 0 L 0.322005 1 M 0.342003 0 L 0.342003 1 M 0.362001 0 L 0.362001 1 M 0.381999 0 L 0.381999 1 M 0.401997 0 L 0.401997 1 M 0.422005 0 L 0.422005 1 M 0.442003 0 L 0.442003 1 M 0.462001 0 L 0.462001 1 M 0.481999 0 L 0.481999 1 M 0.501997 0 L 0.501997 1 M 0.522005 0 L 0.522005 1 M 0.542003 0 L 0.542003 1 M 0.562001 0 L 0.562001 1 M 0.581999 0 L 0.581999 1 M 0.601997 0 L 0.601997 1 M 0.622005 0 L 0.622005 1 M 0.642003 0 L 0.642003 1 M 0.662001 0 L 0.662001 1 M 0.681999 0 L 0.681999 1 M 0.701997 0 L 0.701997 1 M 0.722005 0 L 0.722005 1 M 0.742003 0 L 0.742003 1 M 0.762001 0 L 0.762001 1 M 0.781999 0 L 0.781999 1 M 0.801997 0 L 0.801997 1 M 0.822005 0 L 0.822005 1 M 0.842003 0 L 0.842003 1 M 0.862001 0 L 0.862001 1 M 0.881999 0 L 0.881999 1 M 0.901997 0 L 0.901997 1 M 0.922005 0 L 0.922005 1 M 0.942003 0 L 0.942003 1 M 0.962001 0 L 0.962001 1 M 0.981999 0 L 0.981999 1 " transform="matrix(360,0,0,360,0,0)"/>
            </g>
            </svg>
            ''')
        svg = SVG.parse(q, reify=False)
        print(svg.string_xml())
        # svg.write_xml("myfile.svg")

    def test_write_group(self):
        g = Group()
        self.assertEquals(g.string_xml(), "<g />")

    def test_write_rect(self):
        r = Rect("1in", "1in", "3in", "3in", rx="5%")
        self.assertIn(
            r.string_xml(),
            (
                '<rect rx="0.15in" x="1in" y="1in" ry="0.15in" width="3in" height="3in" />',
                '<rect height="3in" rx="0.15in" ry="0.15in" width="3in" x="1in" y="1in" />',
            ),
        )
        r *= "scale(3)"
        self.assertIn(
            r.string_xml(),
            (
                '<rect rx="0.15in" x="1in" y="1in" ry="0.15in" width="3in" height="3in" transform="matrix(3.000000, 0.000000, 0.000000, 3.000000, 0.000000, 0.000000)" />',
                '<rect height="3in" rx="0.15in" ry="0.15in" transform="matrix(3.000000, 0.000000, 0.000000, 3.000000, 0.000000, 0.000000)" width="3in" x="1in" y="1in" />',
            ),
        )
        r.reify()
        self.assertIn(
            r.string_xml(),
            (
                '<rect rx="0.45in" x="3in" y="3in" ry="0.45in" width="9in" height="9in" />',
                '<rect height="9in" rx="0.45in" ry="0.45in" width="9in" x="3in" y="3in" />',
            ),
        )
        r = Path(r)
        self.assertIn(
            r.string_xml(),
            (
                '<path rx="5%" d="M 3.45,3 L 11.55,3 A 0.45,0.45 0 0,1 12,3.45 L 12,11.55 A 0.45,0.45 0 0,1 11.55,12 L 3.45,12 A 0.45,0.45 0 0,1 3,11.55 L 3,3.45 A 0.45,0.45 0 0,1 3.45,3 Z" />',
                '<path d="M 3.45,3 L 11.55,3 A 0.45,0.45 0 0,1 12,3.45 L 12,11.55 A 0.45,0.45 0 0,1 11.55,12 L 3.45,12 A 0.45,0.45 0 0,1 3,11.55 L 3,3.45 A 0.45,0.45 0 0,1 3.45,3 Z" rx="5%" />',
            ),
        )

    def test_write_path(self):
        r = Path("M0,0zzzz")
        r *= "translate(5,5)"
        self.assertEqual(
            r.string_xml(),
            '<path d="M 0,0 z z z z" transform="matrix(1.000000, 0.000000, 0.000000, 1.000000, 5.000000, 5.000000)" />',
        )
        r.reify()
        self.assertEqual(r.string_xml(), '<path d="M 5,5 z z z z" />')

    def test_write_circle(self):
        c = Circle(r=5, stroke="none", fill="yellow")
        q = SVG.parse(io.StringIO(c.string_xml()))
        self.assertEqual(c, q)

    def test_write_ellipse(self):
        c = Ellipse(rx=3, ry=2, fill="cornflower blue")
        q = SVG.parse(io.StringIO(c.string_xml()))
        self.assertEqual(c, q)

    def test_write_line(self):
        c = SimpleLine(x1=0, x2=10, y1=5, y2=6, id="line", fill="light grey")
        q = SVG.parse(io.StringIO(c.string_xml()))
        self.assertEqual(c, q)

    def test_write_pathlib_issue_227(self):
        """
        Tests pathlib.Path file saving. This is permitted by the xml writer but would crash see issue #227

        This also provides an example of pretty-print off and short_empty_elements off (an xml writer option).

        """
        file1 = "myfile.svg"
        self.addCleanup(os.remove, file1)
        file = pathlib.Path(file1)
        svg = SVG(viewport="0 0 1000 1000", height="10mm", width="10mm")
        svg.append(Rect("10%", "10%", "80%", "80%", fill="red"))
        svg.write_xml(file, pretty=False, short_empty_elements=False)

    def test_write_filename(self):
        """
        Tests filename file saving. This is permitted by the xml writer but would crash see issue #227

        This also provides an example short_empty_elements off, utf-8 encoding.

        """
        file1 = "myfile-f.svg"
        self.addCleanup(os.remove, file1)
        svg = SVG(viewport="0 0 1000 1000", height="10mm", width="10mm")
        svg.append(Rect("10%", "10%", "80%", "80%", fill="red"))
        svg.write_xml(file1, short_empty_elements=False, encoding="utf-8")

    def test_write_filename_svgz(self):
        """
        Tests pathlib.Path file saving. This is permitted by the xml writer but would crash see issue #227

        This also provides an example of xml_declaration=True.

        """
        file1 = "myfile-f.svgz"
        self.addCleanup(os.remove, file1)
        svg = SVG(viewport="0 0 1000 1000", height="10mm", width="10mm")
        svg.append(Rect("10%", "10%", "80%", "80%", fill="red"))
        svg.write_xml(file1, xml_declaration=True)


    # def test_read_write(self):
    #     import glob
    #     for g in glob.glob("*.svg"):
    #         if g.startswith("test-"):
    #             continue
    #         print(g)
    #         try:
    #             svg = SVG.parse(g, transform="translate(1,1)")
    #         except ParseError:
    #             print(f"{g} could not be parsed.")
    #             continue
    #         except ValueError:
    #             continue
    #         svg.write_xml(f"test-{g}.")



import io
import unittest

from svgelements import *


class TestStrokeWidth(unittest.TestCase):
    """Tests the functionality of the Stroke-Width values"""

    def test_viewport_scaling_stroke(self):
        """
        See Issue #199

        The stroke width of both objects here is the same and there is no scaling on the path itself. However, the
        viewport is scaled and applies that scaling to the overall result. Given that these two objects should have
        if drawn in a static fashion the same size stroke. They should not scale based on the viewport since that
        governs a sort of zoom and pan functionality which can be taken as equal to stroke but not in regard to
        vector-effect="non-scaling-stroke".
        """
        q = io.StringIO(
            """<svg version="1.1" xmlns="http://www.w3.org/2000/svg"
            xmlns:xlink="http://www.w3.org/1999/xlink"
            xmlns:ev="http://www.w3.org/2001/xml-events"
            width="335.0mm" height="215.0mm" viewBox="0 0 1266 8130">
            <path d="M 288.706174886,349.064235392 A 113.386,75.5906 0 0,1 175.320348114,424.654786573 A 113.386,75.5906 0 0,1 61.9345213427,349.064235392 A 113.386,75.5906 0 0,1 175.320348114,273.473684211 A 113.386,75.5906 0 0,1 288.706174886,349.064235392 Z" type="elem ellipse" id="meerk40t:15"
            vector-effect="non-scaling-stroke"
            fill-rule="evenodd" stroke="#ff0000" stroke-width="18.897637795275593px" fill="none" />
            <path d="M 302.362204724,151.181102362 A 113.386,75.5906 0 0,1 188.976377953,226.771653543 A 113.386,75.5906 0 0,1 75.5905511811,151.181102362 A 113.386,75.5906 0 0,1 188.976377953,75.5905511811 A 113.386,75.5906 0 0,1 302.362204724,151.181102362 Z" type="elem ellipse" id="meerk40t:18"
            fill-rule="evenodd" stroke="#0000ff" stroke-width="18.897637795275593px" fill="none" />
            </svg>
            """
        )
        m = SVG.parse(q)
        q = list(m.elements())
        self.assertAlmostEqual(q[1].stroke_width, q[2].stroke_width)



import random
import unittest
from random import *

from svgelements import *


def get_random_cubic_bezier():
    return CubicBezier(
        (random() * 50, random() * 50),
        (random() * 50, random() * 50),
        (random() * 50, random() * 50),
        (random() * 50, random() * 50),
    )


class TestElementCubicBezierLength(unittest.TestCase):
    def test_cubic_bezier_length(self):
        n = 100
        error = 0
        for _ in range(n):
            b = get_random_cubic_bezier()
            l1 = b._length_scipy()
            l2 = b._length_default(error=1e-6)
            c = abs(l1 - l2)
            error += c
            self.assertAlmostEqual(l1, l2, places=1)
        print("Average cubic-line error: %g" % (error / n))


class TestElementCubicBezierPoint(unittest.TestCase):
    def test_cubic_bezier_point_start_stop(self):
        import numpy as np

        for _ in range(1000):
            b = get_random_cubic_bezier()
            self.assertEqual(b.start, b.point(0))
            self.assertEqual(b.end, b.point(1))
            self.assertTrue(
                np.all(np.array([list(b.start), list(b.end)]) == b.npoint([0, 1]))
            )

    def test_cubic_bezier_point_implementations_match(self):
        import numpy as np

        for _ in range(1000):
            b = get_random_cubic_bezier()

            pos = np.linspace(0, 1, 100)

            v1 = b.npoint(pos)
            v2 = []
            for i in range(len(pos)):
                v2.append(b.point(pos[i]))

            for p, p1, p2 in zip(pos, v1, v2):
                self.assertEqual(b.point(p), Point(p1))
                self.assertEqual(Point(p1), Point(p2))

    def test_cubic_bounds_issue_214(self):
        cubic = CubicBezier(0, -2 - 3j, -1 - 4j, -3j)
        bbox = cubic.bbox()
        self.assertLess(bbox[1], -3)

    def test_cubic_bounds_issue_214_random(self):
        for i in range(100):
            a = random() * 5
            b = random() * 5
            c = random() * 5
            d = a - 3 * b + 3 * c
            cubic1 = CubicBezier(a, b, c, d)
            bbox1 = cubic1.bbox()
            cubic2 = CubicBezier(a, b, c, d + 1e-11)
            bbox2 = cubic2.bbox()
            for a, b in zip(bbox1, bbox2):
                self.assertAlmostEqual(a, b, delta=1e-5)

    def test_cubic_bounds_issue_220(self):
        p = Path(transform=Matrix(682.657124793113, 0.000000000003, -0.000000000003, 682.657124793113, 257913.248909660178, -507946.354527872754))
        p += CubicBezier(start=Point(-117.139521365,1480.99923469), control1=Point(-41.342266634,1505.62725567), control2=Point(40.3422666342,1505.62725567), end=Point(116.139521365,1480.99923469))
        bounds = p.bbox()
        self.assertNotAlmostEquals(bounds[1], bounds[3], delta=100)



import io
import unittest

from svgelements import *


class TestElementViewbox(unittest.TestCase):

    def test_viewbox_creation(self):
        """Test various ways of creating a viewbox are equal."""
        v1 = Viewbox('0 0 100 100', 'xMid')
        v2 = Viewbox(viewBox="0 0 100 100", preserve_aspect_ratio="xMid")
        v3 = Viewbox(x=0, y=0, width=100, height=100, preserveAspectRatio="xMid")
        v4 = Viewbox(v1)
        v5 = Viewbox({"x":0, "y":0, "width":100, "height":100, "preserveAspectRatio":"xMid"})
        self.assertEqual(v1, v2)
        self.assertEqual(v1, v3)
        self.assertEqual(v1, v4)
        self.assertEqual(v1, v5)
        self.assertEqual(v2, v3)
        self.assertEqual(v2, v4)
        self.assertEqual(v2, v5)
        self.assertEqual(v3, v4)
        self.assertEqual(v3, v5)
        self.assertEqual(v4, v5)

    def test_viewbox_incomplete_none(self):
        """
        Test viewboxes based on incomplete information.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg/>''')
        m = SVG.parse(q)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 1000)
        self.assertEqual(m.height, 1000)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 500)
        self.assertEqual(m.height, 500)

    def test_viewbox_incomplete_height(self):
        """
        Test viewboxes based on incomplete information, only height.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg height="200"/>''')
        m = SVG.parse(q)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 1000)
        self.assertEqual(m.height, 200)
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg height="200"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 500)
        self.assertEqual(m.height, 200)

    def test_viewbox_incomplete_width(self):
        """
        Test viewboxes based on incomplete information, only width.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg width="200"/>''')
        m = SVG.parse(q)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 200)
        self.assertEqual(m.height, 1000)
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg width="200"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 200)
        self.assertEqual(m.height, 500)

    def test_viewbox_incomplete_dims(self):
        """
        Test viewboxes based on incomplete information, only dims.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg width="200" height="200"/>''')
        m = SVG.parse(q)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 200)
        self.assertEqual(m.height, 200)
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg width="200" height="200"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(m.viewbox_transform, '')
        self.assertEqual(m.width, 200)
        self.assertEqual(m.height, 200)

    def test_viewbox_incomplete_viewbox(self):
        """
        Test viewboxes based on incomplete information, only viewbox.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0, 0, 100, 100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(1)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 100)
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0, 0, 100, 100"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(5)')
        self.assertEqual(m.width, 500)
        self.assertEqual(m.height, 500)

    def test_viewbox_incomplete_height_viewbox(self):
        """
        Test viewboxes based on incomplete information, only height and viewbox.
        """
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0, 0, 100, 100" height="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), '')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 100)
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0, 0, 100, 100" height="100"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(1) translateX(200)')
        self.assertEqual(m.width, 500)
        self.assertEqual(m.height, 100)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0, 0, 100, 100" height="200" width="200"/>''')
        m = SVG.parse(q, width=500, height=500)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(2)')
        self.assertEqual(m.width, 200)
        self.assertEqual(m.height, 200)

    def test_viewbox_aspect_ratio_xMinMax(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMid" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateX(100)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMin" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateX(0)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMax" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateX(200)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

    def test_viewbox_aspect_ratio_xMinMaxSlice(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMid slice" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMin slice" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="xMax slice" viewBox="0 0 100 100" height="100" width="300"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 300)
        self.assertEqual(m.height, 100)

    def test_viewbox_aspect_ratio_yMinMax(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMid" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateY(100)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMin" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateY(0)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMax" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'translateY(200)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

    def test_viewbox_aspect_ratio_yMinMaxSlice(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMid slice" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMin slice" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg preserveAspectRatio="yMax slice" viewBox="0 0 100 100" height="300" width="100"/>''')
        m = SVG.parse(q)
        self.assertEqual(Matrix(m.viewbox_transform), 'scale(3)')
        self.assertEqual(m.width, 100)
        self.assertEqual(m.height, 300)

    def test_viewbox_simple(self):
        r = Rect(0, 0, 100, 100)
        v = Viewbox({'viewBox': '0 0 100 100'})
        self.assertEqual(v.transform(r), '')

    def test_viewbox_issue_228(self):
        self.assertIn(
            SVG(viewBox="0 0 10 10", width="10mm", height="10mm").string_xml(),
            (
                """<svg viewBox="0 0 10 10" width="10mm" height="10mm" version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:ev="http://www.w3.org/2001/xml-events" />""",
                """<svg height="10mm" version="1.1" viewBox="0 0 10 10" width="10mm" xmlns="http://www.w3.org/2000/svg" xmlns:ev="http://www.w3.org/2001/xml-events" xmlns:xlink="http://www.w3.org/1999/xlink" />""",  # Python 3.6
            ),
        )

    def test_issue_228b(self):
        svg = SVG(viewBox="0 0 10 10", width="10mm", height="10mm")
        svg.append(Rect(x="1mm", y="1mm", width="5mm", height="5mm", rx="0.5mm", stroke="red"))
        svg.append(Circle(cx="5mm", cy="5mm", r="0.5mm", stroke="blue"))
        svg.append(Ellipse(cx="5mm", cy="5mm", rx="0.5mm", ry="0.8mm", stroke="lime"))
        svg.append(SimpleLine(x1="5mm", y1="5em", x2="10%", y2="15%", stroke="gray"))
        svg.append(Polygon(5, 10, 20, 30, 40, 7))
        svg.append(Path("M10,10z", stroke="yellow"))
        print(svg.string_xml())

    def test_issue_228c(self):
        rect = Rect(x="1mm", y="1mm", width="5mm", height="5mm", rx="0.5mm", stroke="red")
        print(rect.length())

    def test_viewbox_scale(self):
        r = Rect(0, 0, 200, 200)
        v = Viewbox('0 0 100 100')
        self.assertEqual(v.transform(r), 'scale(2, 2)')

    def test_viewbox_translate(self):
        r = Rect(0, 0, 100, 100)
        v = Viewbox(Viewbox('-50 -50 100 100'))
        self.assertEqual(v.transform(r), 'translate(50, 50)')

    def test_viewbox_parse_empty(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg>
                        </svg>''')
        m = SVG.parse(q)
        q = list(m.elements())
        self.assertEqual(len(q), 1)
        self.assertEqual(None, m.viewbox)

    def test_viewbox_parse_100(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="0 0 100 100">
                        </svg>''')
        m = SVG.parse(q, width=100, height=100)
        q = list(m.elements())
        self.assertEqual(len(q), 1)
        self.assertEqual(Matrix(m.viewbox_transform), Matrix.identity())

    def test_viewbox_parse_translate(self):
        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg viewBox="-1 -1 100 100" width="100" height="100">
                        </svg>''')
        m = SVG.parse(q)
        q = list(m.elements())
        self.assertEqual(len(q), 1)
        self.assertEqual(Matrix(m.viewbox_transform), Matrix.translate(1, 1))



import unittest

from svgelements import *


class TestElementLinear(unittest.TestCase):

    def test_linear_nearest(self):
        line = Line((0,0),(5,0))
        r = line.closest_segment_point((17,0))
        self.assertEqual(r, (5,0))
        r = line.closest_segment_point((2, 2))
        self.assertEqual(r, (2, 0))


class TestBoundingBox(unittest.TestCase):

    def test_linear_bbox(self):
        line = Line((0,0), (5,0))
        r = line.bbox()
        self.assertEqual(r, (0, 0, 5, 0))

    def test_qbezier_bbox(self):
        line = QuadraticBezier((0,0), (2,2), (5,0))
        r = line.bbox()
        self.assertEqual(r, (0, 0, 5, 1))

    def test_cbezier_bbox(self):
        line = CubicBezier((0,0), (2,2), (2,-2), (5,0))
        r = line.bbox()
        for z in zip(r, (0.0, -0.5773502691896257, 5.0, 0.5773502691896257)):
            self.assertAlmostEqual(*z)

    def test_arc_bbox(self):
        line = Arc((0,0), (5,0), control=(2.5, 2.5))
        r = line.bbox()
        for z in zip(r, (0.0, 0, 5.0, 2.5)):
            self.assertAlmostEqual(*z)

    def test_null_arc_bbox(self):
        self.assertEqual(Path("M0,0A0,0 0 0 0 0,0z").bbox(), (0,0,0,0))


class TestArcControlPoints(unittest.TestCase):

    def test_coincident_end_arc(self):
        """
        Tests the creation of a control point with a coincident start and end.
        """
        arc = Arc(start=(0,0), control=(50,0), end=(0,0))
        self.assertAlmostEqual(arc.rx, 25)

    def test_linear_arc(self):
        """
        Colinear Arcs should raise value errors.
        """
        arc_vertical = Arc(start=(0, 0), control=(25, 0), end=(50, 0))
        # print(arc_vertical)
        arc_horizontal = Arc(start=(0, 0), control=(0, 25), end=(0, 50))
        # print(arc_horizontal)



import unittest
from random import *

from svgelements import *


def get_random_cubic_bezier():
    return CubicBezier((random() * 50, random() * 50), (random() * 50, random() * 50),
                       (random() * 50, random() * 50), (random() * 50, random() * 50))


class TestElementApproximation(unittest.TestCase):

    def test_cubic_bezier_arc_approximation(self):
        n = 50
        for _ in range(n):
            b = get_random_cubic_bezier()
            path = Move(b.start) + Path([b])
            path2 = Path(path)
            path2.approximate_bezier_with_circular_arcs(error=0.001)
            path2.approximate_arcs_with_cubics(error=0.001)



import unittest
from operator import itemgetter

from svgelements import *

TOL = 1e-4  # default for tests that don't specify a `delta` or `places`


class TestElementIntersections(unittest.TestCase):

    def test_intersect(self):
        """
        test that `some_seg.intersect(another_seg)` will produce properly
        ordered tuples, i.e. the first element in each tuple refers to
        `some_seg` and the second element refers to `another_seg`.
        Also tests that the correct number of intersections is found.

        * This test adapted from svgpathtools
        """
        a = Line(0 + 200j, 300 + 200j)
        b = QuadraticBezier(40 + 150j, 70 + 200j, 210 + 300j)
        c = CubicBezier(60 + 150j, 40 + 200j, 120 + 250j, 200 + 160j)
        d = Arc(70 + 150j, 50 + 100j, 0, 0, 0, 200 + 100j)
        segdict = {'line': a, "quadratic": b, 'cubic': c, 'arc': d}

        # test each segment type against each other type
        for x, y in [(x, y) for x in segdict for y in segdict]:
            if x == y:
                continue
            x = segdict[x]
            y = segdict[y]
            xiy = sorted(x.intersect(y))
            yix = sorted(y.intersect(x), key=itemgetter(1))
            for xy, yx in zip(xiy, yix):
                self.assertAlmostEqual(xy[0], yx[1], delta=TOL)
                self.assertAlmostEqual(xy[1], yx[0], delta=TOL)
                self.assertAlmostEqual(x.point(xy[0]), y.point(yx[0]), delta=TOL)
            self.assertTrue(len(xiy) == len(yix))

        # test each segment against another segment of same type
        for x in segdict:
            count = 1
            if x == "arc":
                count = 2
            x = segdict[x]
            mid = x.point(0.5)
            y = x * Matrix(f"translate(5,0) rotate(90, {mid.x}, {mid.y})")
            xiy = sorted(x.intersect(y))
            yix = sorted(y.intersect(x), key=itemgetter(1))
            for xy, yx in zip(xiy, yix):
                self.assertAlmostEqual(xy[0], yx[1], delta=TOL)
                self.assertAlmostEqual(xy[1], yx[0], delta=TOL)
                self.assertAlmostEqual(x.point(xy[0]), y.point(yx[0]), delta=TOL)
            self.assertTrue(len(xiy) == len(yix))

            self.assertTrue(len(xiy) == count)
            self.assertTrue(len(yix) == count)



import unittest
import io

from svgelements import *


class TestElementBbox(unittest.TestCase):

    def test_bbox_rect(self):
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10"
        }
        e = Rect(values)
        self.assertEqual(e.bbox(), (50,51,70,61))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))

    def test_bbox_rect_stroke(self):
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10",
            'stroke-width': "5",
            'stroke': 'red',
        }
        e = Rect(values)
        self.assertEqual(e.bbox(), (50, 51, 70, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            50-(5./2.),
            51-(5./2.),
            70+(5./2.),
            61+(5./2.)
        ))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 - (5. / 2.),
            51 - (5. / 2.),
            72 + (5. / 2.),
            61 + (5. / 2.)
        ))
        e *= "scale(2)"
        self.assertEqual(e.bbox(), (52 * 2, 51 * 2, 72 * 2, 61 * 2))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 * 2 - 5.,
            51 * 2 - 5.,
            72 * 2 + 5.,
            61 * 2 + 5.
        ))
        self.assertEqual(e.bbox(transformed=False), (50, 51, 70, 61))
        self.assertEqual(e.bbox(transformed=False, with_stroke=True), (
            50 - (5. / 2.),
            51 - (5. / 2.),
            70 + (5. / 2.),
            61 + (5. / 2.)
        ))

    def test_bbox_path(self):
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10"
        }
        e = Path(Rect(values))
        self.assertEqual(e.bbox(), (50,51,70,61))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))

    def test_bbox_path_stroke(self):
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10",
            'stroke-width': "5",
            'stroke': 'red',
        }
        e = Path(Rect(values))
        self.assertEqual(e.bbox(), (50, 51, 70, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            50-(5./2.),
            51-(5./2.),
            70+(5./2.),
            61+(5./2.)
        ))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 - (5. / 2.),
            51 - (5. / 2.),
            72 + (5. / 2.),
            61 + (5. / 2.)
        ))
        e *= "scale(2)"
        self.assertEqual(e.bbox(), (52 * 2, 51 * 2, 72 * 2, 61 * 2))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 * 2 - 5.,
            51 * 2 - 5.,
            72 * 2 + 5.,
            61 * 2 + 5.
        ))
        self.assertEqual(e.bbox(transformed=False), (50, 51, 70, 61))
        self.assertEqual(e.bbox(transformed=False, with_stroke=True), (
            50 - (5. / 2.),
            51 - (5. / 2.),
            70 + (5. / 2.),
            61 + (5. / 2.)
        ))

    def test_bbox_path_stroke_none(self):
        """
        Same as test_bbox_path_stroke but stroke is set to none, so the bbox should not change.
        """
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10",
            'stroke-width': "5",
            'stroke': "none",
        }
        e = Path(Rect(values))
        self.assertEqual(e.bbox(), (50, 51, 70, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            50,
            51,
            70,
            61
        ))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            52,
            51,
            72,
            61
        ))
        e *= "scale(2)"
        self.assertEqual(e.bbox(), (52 * 2, 51 * 2, 72 * 2, 61 * 2))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 * 2,
            51 * 2,
            72 * 2,
            61 * 2
        ))
        self.assertEqual(e.bbox(transformed=False), (50, 51, 70, 61))
        self.assertEqual(e.bbox(transformed=False, with_stroke=True), (
            50,
            51,
            70,
            61
        ))

    def test_bbox_path_stroke_unset(self):
        """
        Same as test_bbox_path_stroke but the stroke is unset and thus shouldn't contribute to the bbox even if
        with_stroke is set.
        """
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10",
            'stroke-width': "5",
        }
        e = Path(Rect(values))
        self.assertEqual(e.bbox(), (50, 51, 70, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            50,
            51,
            70,
            61
        ))
        e *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            52,
            51,
            72,
            61
        ))
        e *= "scale(2)"
        self.assertEqual(e.bbox(), (52 * 2, 51 * 2, 72 * 2, 61 * 2))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 * 2,
            51 * 2,
            72 * 2,
            61 * 2
        ))
        self.assertEqual(e.bbox(transformed=False), (50, 51, 70, 61))
        self.assertEqual(e.bbox(transformed=False, with_stroke=True), (
            50,
            51,
            70,
            61
        ))

    def test_bbox_subpath(self):
        p = Path("M 10,100 H 20 V 80 H 10 Z m 10,-90 H 60 V 70 H 20 Z")
        e = p.subpath(1)
        self.assertEqual(e.bbox(), (20, 10, 60, 70))
        e *= "translate(5)"
        self.assertEqual(e.bbox(), (25, 10, 65, 70))

    def test_bbox_move_subpath2(self):
        p = Path("M 0,0 Z m 100,100 h 20 v 20 h -20 Z")
        e = p.subpath(1)
        self.assertEqual(e.bbox(), (100, 100, 120, 120))

    def test_bbox_subpath_stroke(self):
        values = {
            'tag': 'rect',
            'rx': "4",
            'ry': "2",
            'x': "50",
            'y': "51",
            'width': "20",
            'height': "10",
            'stroke-width': "5",
            'stroke': 'red',
        }
        p = Path(Rect(values))
        e = p.subpath(0)
        self.assertEqual(e.bbox(), (50, 51, 70, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            50-(5./2.),
            51-(5./2.),
            70+(5./2.),
            61+(5./2.)
        ))
        p *= "translate(2)"
        self.assertEqual(e.bbox(), (52, 51, 72, 61))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 - (5. / 2.),
            51 - (5. / 2.),
            72 + (5. / 2.),
            61 + (5. / 2.)
        ))
        p *= "scale(2)"
        self.assertEqual(e.bbox(), (52 * 2, 51 * 2, 72 * 2, 61 * 2))
        self.assertEqual(e.bbox(with_stroke=True), (
            52 * 2 - 5.,
            51 * 2 - 5.,
            72 * 2 + 5.,
            61 * 2 + 5.
        ))
        self.assertEqual(e.bbox(transformed=False), (50, 51, 70, 61))
        self.assertEqual(e.bbox(transformed=False, with_stroke=True), (
            50 - (5. / 2.),
            51 - (5. / 2.),
            70 + (5. / 2.),
            61 + (5. / 2.)
        ))

    def test_bbox_rotated_circle(self):
        # Rotation of circle must not affect it's bounding box
        c = Circle(cx=0, cy=0, r=1, transform="rotate(45)")
        (xmin, ymin, xmax, ymax) = c.bbox()
        self.assertAlmostEqual(-1, xmin)
        self.assertAlmostEqual(-1, ymin)
        self.assertAlmostEqual( 1, xmax)
        self.assertAlmostEqual( 1, ymax)

    def test_bbox_svg_with_rotated_circle(self):
        # Rotation of circle within group must not affect it's bounding box
        q = io.StringIO(
            u'''<?xml version="1.0" encoding="utf-8" ?>
            <svg>
              <circle cx="0" cy="0" r="1" transform="rotate(45)"/>
            </svg>
            '''
        )
        svg = SVG.parse(q)
        (xmin, ymin, xmax, ymax) = svg.bbox()
        self.assertAlmostEqual(-1, xmin)
        self.assertAlmostEqual(-1, ymin)
        self.assertAlmostEqual( 1, xmax)
        self.assertAlmostEqual( 1, ymax)

    def test_bbox_translated_circle(self):
        c = Circle(cx=0, cy=0, r=1, transform="translate(-1,-1)")
        (xmin, ymin, xmax, ymax) = c.bbox()
        self.assertAlmostEqual(-2, xmin)
        self.assertAlmostEqual(-2, ymin)
        self.assertAlmostEqual( 0, xmax)
        self.assertAlmostEqual( 0, ymax)

    def test_bbox_svg_with_translated_group_with_circle(self):
        # Translation of nested group must be applied correctly
        q = io.StringIO(
            u'''<?xml version="1.0" encoding="utf-8" ?>
            <svg>
              <g transform="translate(-1,-1)">
                <circle cx="0" cy="0" r="1"/>
              </g>
            </svg>
            '''
        )
        svg = SVG.parse(q)
        (xmin, ymin, xmax, ymax) = svg.bbox()
        self.assertAlmostEqual(-2, xmin)
        self.assertAlmostEqual(-2, ymin)
        self.assertAlmostEqual( 0, xmax)
        self.assertAlmostEqual( 0, ymax)


    def test_issue_104(self):
        """Testing Issue 104 rotated bbox"""
        rect = Rect(10,10,10,10)
        rect *= "rotate(45deg)"
        self.assertEqual(rect.bbox(), abs(rect).bbox())

        circ = Circle(5,5,10)
        circ *= "rotate(45deg)"
        self.assertEqual(circ.bbox(), abs(circ).bbox())

        path = Path("M0 0 100,100")
        path *= "rotate(45deg)"
        self.assertEqual(path.bbox(), abs(path).bbox())

        path = Path("M0 0q100,100 200,200z")
        path *= "rotate(45deg)"
        self.assertEqual(path.bbox(), abs(path).bbox())

        path = Path("M0 20c0,128 94,94 200,200z")
        path *= "rotate(45deg)"
        self.assertEqual(path.bbox(), abs(path).bbox())

        q = io.StringIO(u'''<?xml version="1.0" encoding="utf-8" ?>
                        <svg>
                        <path d="M0 36c0 128.082 96 214 251 214c153.639 0 293 -37 293 -264c0 -220 -158.05 -325.976 -254 -391c-121 -82 -200.248 -117.634 -339 -171c-5.93907 -2.28437 -11.3166 -3.27496 -15.9795 -3.27496c-11.9348 0 -19.1871 6.48977 -19.1871 14.3854 c0 7.04844 5.77949 15.2172 19.1666 20.8896c118 50 225 86 316 200c91 113 125 205.913 125 350c0 135 -33 224 -145 224c-46.615 0 -77.452 -12.593 -112 -44c-8.09234 -7.35667 -11.3121 -13.0296 -11.3121 -17.4022c0 -13.4585 30.5027 -14.5978 43.3121 -14.5978 c69 0 123 -64.8867 123 -136c0 -75 -48 -132 -139 -132c-79.1582 0 -135 74 -135 150zM572 128c0 35 29 64 64 64s63 -29 63 -64s-28 -63 -63 -63s-64 28 -64 63zM572 -131c0 35 29 64 64 64s63 -29 63 -64s-28 -63 -63 -63s-64 28 -64 63z" transform="rotate(45deg)"/>
                        </svg>''')
        m = SVG.parse(q, reify=False)
        p0 = m[0]
        p1 = abs(p0)
        self.assertEqual(p0, p1)
        self.assertEqual(p0.bbox(), p1.bbox())
